from __future__ import annotations

from typing import TYPE_CHECKING

from flask_pancake import RAW_FALSE, RAW_TRUE, Switch

if TYPE_CHECKING:
    from flask import Flask


def test_switch(app: Flask):
    off = Switch("default_off", False)
    on = Switch("default_on", True)

    assert app.extensions["redis"].get("SWITCH:DEFAULT_OFF") is None
    assert app.extensions["redis"].get("SWITCH:DEFAULT_ON") is None

    assert off.is_active() is False
    assert on.is_active() is True

    assert app.extensions["redis"].get("SWITCH:DEFAULT_OFF") == RAW_FALSE
    assert app.extensions["redis"].get("SWITCH:DEFAULT_ON") == RAW_TRUE

    off.enable()
    on.disable()

    assert app.extensions["redis"].get("SWITCH:DEFAULT_OFF") == RAW_TRUE
    assert app.extensions["redis"].get("SWITCH:DEFAULT_ON") == RAW_FALSE

    assert off.is_active() is True
    assert on.is_active() is False

    off.clear()
    assert app.extensions["redis"].get("SWITCH:DEFAULT_OFF") is None



import pytest
from flask import Flask
from flask_redis import FlaskRedis

from flask_pancake import EXTENSION_NAME, FlaskPancake


@pytest.fixture
def _app():
    app = Flask(__name__)
    app.config["TESTING"] = True
    FlaskRedis(app)
    FlaskPancake(app)
    yield app


@pytest.fixture
def app(_app):
    with _app.app_context():
        yield _app


@pytest.fixture
def client(app):
    with app.test_client() as client:
        yield client


@pytest.fixture(autouse=True)
def flask_pancake_cleanup(app: Flask):
    def inner():
        flask_pancake: FlaskPancake = app.extensions[EXTENSION_NAME]
        for flag in flask_pancake.flags.values():
            flag.clear()
            flag.clear_all_users()
        for switch in flask_pancake.switches.values():
            switch.clear()
        for sample in flask_pancake.samples.values():
            sample.clear()

    inner()
    yield
    inner()



from __future__ import annotations

import uuid
from typing import TYPE_CHECKING

from flask_pancake import EXTENSION_NAME, RAW_FALSE, RAW_TRUE, Flag

if TYPE_CHECKING:
    from flask import Flask


def test_flag(app: Flask):
    uid = str(uuid.uuid4())
    app.extensions[EXTENSION_NAME].get_user_id_func = lambda: uid

    off = Flag("default_off", False)
    on = Flag("default_on", True)

    assert app.extensions["redis"].get("FLAG:DEFAULT_OFF") is None
    assert app.extensions["redis"].get("FLAG:DEFAULT_ON") is None

    assert off.is_active() is False
    assert on.is_active() is True

    assert app.extensions["redis"].get("FLAG:DEFAULT_OFF") == RAW_FALSE
    assert app.extensions["redis"].get("FLAG:DEFAULT_ON") == RAW_TRUE

    off.enable()
    on.disable()

    assert app.extensions["redis"].get("FLAG:DEFAULT_OFF") == RAW_TRUE
    assert app.extensions["redis"].get("FLAG:DEFAULT_ON") == RAW_FALSE

    assert off.is_active() is True
    assert on.is_active() is False


def test_flag_user(app: Flask):
    uid1 = str(uuid.uuid4())
    uid2 = str(uuid.uuid4())
    app.extensions[EXTENSION_NAME].get_user_id_func = lambda: uid1

    feature = Flag("FEATURE", False)

    assert app.extensions["redis"].get("FLAG:FEATURE") is None

    assert feature.is_active() is False
    feature.enable_user()
    assert feature.is_active() is True

    assert app.extensions["redis"].get("FLAG:FEATURE") == RAW_FALSE
    assert app.extensions["redis"].get(f"FLAG:user:{uid1}:FEATURE") == RAW_TRUE

    app.extensions[EXTENSION_NAME].get_user_id_func = lambda: uid2
    assert feature.is_active() is False

    feature.enable()
    assert app.extensions["redis"].get("FLAG:FEATURE") == RAW_TRUE

    assert feature.is_active() is True

    feature.disable_user()

    assert feature.is_active() is False
    assert app.extensions["redis"].get(f"FLAG:user:{uid2}:FEATURE") == RAW_FALSE

    feature.clear()
    assert app.extensions["redis"].get("SAMPLE:FEATURE") is None



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest import mock

from flask_pancake import Sample

if TYPE_CHECKING:
    from flask import Flask


def test_sample(app: Flask):
    feature = Sample("feature", 42)

    assert app.extensions["redis"].get("SAMPLE:FEATURE") is None

    with mock.patch("random.uniform", return_value=42.00001):
        assert feature.is_active() is False
    assert app.extensions["redis"].get("SAMPLE:FEATURE") == b"42"
    with mock.patch("random.uniform", return_value=42):
        assert feature.is_active() is True

    feature.clear()
    assert app.extensions["redis"].get("SAMPLE:FEATURE") is None



from __future__ import annotations

import abc
import random
from typing import TYPE_CHECKING, Callable, Dict, Generic, Optional, TypeVar, Union

from cached_property import cached_property
from flask import current_app

from .utils import import_from_string

if TYPE_CHECKING:
    from flask import Flask
    from flask_redis import FlaskRedis


DEFAULT_TYPE = TypeVar("DEFAULT_TYPE")
EXTENSION_NAME = "pancake"

RAW_FALSE = b"0"
RAW_TRUE = b"1"


_FLAGS: Dict[str, Flag] = {}
_SWITCHES: Dict[str, Switch] = {}
_SAMPLES: Dict[str, Sample] = {}


class FlaskPancake:
    def __init__(
        self,
        app: Flask = None,
        *,
        redis_extension_name: str = "redis",
        get_user_id_func: Optional[Union[str, Callable[[], str]]] = None,
    ) -> None:
        self.redis_extension_name = redis_extension_name
        self._get_user_id_func = get_user_id_func

        self.app = app
        if app is not None:
            self.init_app(app)

    def init_app(self, app: Flask) -> None:
        app.extensions[EXTENSION_NAME] = self

    @cached_property
    def get_user_id_func(self) -> Optional[Callable[[], str]]:
        if isinstance(self._get_user_id_func, str):
            return import_from_string(self._get_user_id_func)
        return self._get_user_id_func

    @property
    def flags(self) -> Dict[str, Flag]:
        return _FLAGS

    @property
    def switches(self) -> Dict[str, Switch]:
        return _SWITCHES

    @property
    def samples(self) -> Dict[str, Sample]:
        return _SAMPLES


class AbstractFlag(abc.ABC, Generic[DEFAULT_TYPE]):
    name: str

    def __init__(self, name: str, default: DEFAULT_TYPE) -> None:
        self.name = name
        self.set_default(default)

        if isinstance(self, Flag):
            _FLAGS[name.upper()] = self
        elif isinstance(self, Switch):
            _SWITCHES[name.upper()] = self
        elif isinstance(self, Sample):
            _SAMPLES[name.upper()] = self

    def set_default(self, default: DEFAULT_TYPE) -> None:
        self.default = default

    @cached_property
    def ext(self) -> FlaskPancake:
        return current_app.extensions[EXTENSION_NAME]

    @cached_property
    def _redis_client(self) -> FlaskRedis:
        return current_app.extensions[self.ext.redis_extension_name]

    @cached_property
    def key(self) -> str:
        return f"{self.__class__.__name__.upper()}:{self.name.upper()}"

    @abc.abstractmethod
    def is_active(self) -> bool:
        ...

    def clear(self) -> None:
        self._redis_client.delete(self.key)


class BaseFlag(AbstractFlag[bool], abc.ABC):
    def set_default(self, default: bool) -> None:
        if int(default) not in {0, 1}:
            raise ValueError(
                f"Default value for switch {self.name} must be True or False."
            )
        super().set_default(default)

    def is_active(self) -> bool:
        self._redis_client.setnx(self.key, int(self.default))
        return self._redis_client.get(self.key) == RAW_TRUE

    def disable(self) -> None:
        self._redis_client.set(self.key, 0)

    def enable(self) -> None:
        self._redis_client.set(self.key, 1)


class Flag(BaseFlag):
    """
    A feature flag.

    Flags are active (or not) on a per-request / user basis.
    """

    @property
    def user_key(self) -> Optional[str]:
        get_user_id_func = self.ext.get_user_id_func
        if get_user_id_func is None:
            raise RuntimeError(
                "No get_user_id_func defined on FlaskPancake. If you don't "
                "have users in your application and want a global flag to "
                "turn things on and off, use a `Switch` instead."
            )
        uid = get_user_id_func()
        if uid is None:
            return None
        return f"{self.__class__.__name__.upper()}:user:{uid}:{self.name.upper()}"

    def is_active(self) -> bool:
        user_key = self.user_key
        if user_key is not None:
            value = self._redis_client.get(user_key)
            if value == RAW_TRUE:
                return True
            elif value == RAW_FALSE:
                return False

        return super().is_active()

    @cached_property
    def _track_key_users(self):
        return f"{self.__class__.__name__.upper()}:track:users"

    def _track_user(self, key):
        self._redis_client.sadd(self._track_key_users, key)

    def clear_user(self) -> None:
        uk = self.user_key
        self._redis_client.delete(uk)
        self._redis_client.srem(self._track_key_users, uk)

    def clear_all_users(self) -> None:
        keys = self._redis_client.smembers(self._track_key_users)
        if keys:
            self._redis_client.delete(*keys)

    def disable_user(self) -> None:
        uk = self.user_key
        self._track_user(uk)
        self._redis_client.set(uk, 0)

    def enable_user(self) -> None:
        uk = self.user_key
        self._track_user(uk)
        self._redis_client.set(uk, 1)


class Switch(BaseFlag):
    """
    A feature switch.

    Switches are active or inactive, globally.
    """


class Sample(AbstractFlag[float]):
    """
    A sample of users.

    A sample is active some percentage of the time, but is not connected to users
    or requests.
    """

    def set_default(self, default: float) -> None:
        if not (0 <= default <= 100):
            raise ValueError(
                f"Default value for sample {self.name} must be in the range [0, 100]."
            )
        super().set_default(default)

    def is_active(self) -> bool:
        self._redis_client.setnx(self.key, self.default)
        value = self._redis_client.get(self.key)
        return random.uniform(0, 100) <= float(value)

    def set(self, value: float) -> None:
        if not (0 <= value <= 100):
            raise ValueError(
                f"Value for sample {self.name} must be in the range [0, 100]."
            )
        self._redis_client.set(self.key, value)



import importlib
from typing import Any


def import_from_string(fqn: str) -> Any:
    module_name, _, attr = fqn.partition(":")
    module = importlib.import_module(module_name)
    if attr:
        return getattr(module, attr)
    return module



from flask import Flask
from flask_redis import FlaskRedis

from flask_pancake import FlaskPancake, Switch

app = Flask(__name__)
pancake = FlaskPancake(app)
redis = FlaskRedis(app)

SWITCH_FEATURE = Switch("FEATURE", default=False)


@app.route("/")
def index():
    if SWITCH_FEATURE.is_active():
        return (
            "Hello World!<br>It worked!<br>Now, try to set the key "
            "<code>SWITCH:FEATURE</code> back to <code>0</code> or delete the "
            "key in Redis with <code>del SWITCH:FEATURE</code>. "
            "Then refresh the page again.",
            200,
        )
    else:
        return (
            "Not found<br>Try setting the key <code>SWITCH:FEATURE</code> "
            "in Redis to <code>1</code> using <code>set SWITCH:FEATURE 1</code>. "
            "Refresh the page afterwards.",
            404,
        )



import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="flask-pancake",
    author="Markus Holtermann",
    author_email="info@markusholtermann.eu",
    description="Feature Flagging for Flask",
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MarkusH/flask-pancake",
    packages=setuptools.find_packages(
        exclude=["*.tests", "*.tests.*", "tests.*", "tests"],
    ),
    install_requires=["flask>=1.0", "flask-redis>=0.4.0", "cached-property>=1.5,<2"],
    extras_require={"testing": ["pytest>=5.3,<5.4"]},
    setup_requires=["setuptools_scm>=3.4.2,<4"],
    use_scm_version=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: Flask",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
    ],
    python_requires=">=3.7",
)
